#include "Distack.h"

//************************************************

#ifndef OMNET_SIMULATION

int main (int argc, char* argv[])
{
	string configfile;
	
	if (argc == 1) {
		
		#ifdef WIN32
			#ifdef _DEBUG
				configfile = "..\\..\\settings.xml";
			#else
				configfile = "settings.xml";
			#endif
		#else
			configfile = "./settings.xml";
		#endif

	} else if (argc == 2) {
		configfile = argv[1];
	} else {
		cout << "Invalid parameters" << std::endl;	
		cout << "Usage: distack settings.xml" << std::endl;
		return false;
	} 

	cout 	<< "-------------------" << std::endl
		<< "Distack v" << VERSION << " --- distributed attack detection ---" << std::endl
		<< "-------------------" << std::endl;

	Distack distack (configfile);

	if (distack.startup ()) {
		getchar			 ();	
		distack.shutdown ();
	}

	getchar ();		
	return 0;
}

#endif // OMNET_SIMULATION

//************************************************

#ifdef OMNET_SIMULATION

DistackOmnetModule::DistackOmnetModule ()
:	omnetInterface	(NULL),
	distack		("settings.xml")
{
}

DistackOmnetModule::~DistackOmnetModule ()
{
	distack.shutdown ();
}

int DistackOmnetModule::numInitStages () const
{
	return 4;
}

void DistackOmnetModule::initialize (int stages)
{
	if (stages < 3) return;

	distack.setDistackModule (this);
	distack.startup			 ();

	NetworkInterface* iface	= distack.getNetworkManager()->getNetworkInterface ();
	omnetInterface = dynamic_cast<NetworkInterfaceOmnetpp*> (iface);
	assert (omnetInterface != NULL);

	string loctest(setlocale(LC_NUMERIC, NULL));
	if(loctest != "en_US.UTF-8") {
		bool result = setlocale(LC_NUMERIC, "en_US.UTF-8");
		if(!result) cout << "libdistack init WARNING: could not change locale settings, simulation may crash" << endl;
	}
}

void DistackOmnetModule::handleMessage	(cMessage* msg)
{
	if (msg->isSelfMessage ()) {

		//
		// handle timer message (self-messages)
		//

		boost::mutex::scoped_lock scopedLock (timerListMutex);
		
		OMNET_TIMER_LIST_ITERATOR i	   = timerList.begin ();
		OMNET_TIMER_LIST_ITERATOR iend = timerList.end   ();

		//
		// inject the timer message to each registered
		// timer. the timer which matches the id will return true
		//
			
		for ( ; i != iend; i++) {
			
			if ((*i)->getTimerId () == msg->kind()) {
				if ((*i)->receiveMessage(msg) == true) 
					break;
			}

		} // for ( ; i != iend; i++)

	} else {

		//
		// handle network messages
		//

		if (omnetInterface != NULL)
			omnetInterface->handleMessage (msg);
	
	}

	//
	// flush the messaging system. otherwise it does not
	// get enough CPU under an OMNET_SIMULATION and will
	// not deliver the messages in a timely fashion
	//

	distack.getModuleManager()->getMessagingSystem().flush ();
}

void DistackOmnetModule::finish ()
{
}

void DistackOmnetModule::registerTimer (Timer* timer)
{
	//
	// if the timer is not yet in the list, register it
	//

	boost::mutex::scoped_lock scopedLock (timerListMutex);
	OMNET_TIMER_LIST_ITERATOR item = find (timerList.begin(), timerList.end(), timer);

	if (item == timerList.end()) {
		timerList.push_back		(timer);
		timer->setDistackModule (this);
	}
}

void DistackOmnetModule::unregisterTimer (Timer* timer)
{
	//
	// if the timer is registered, unregister it
	//

	boost::mutex::scoped_lock scopedLock (timerListMutex);
	OMNET_TIMER_LIST_ITERATOR item = find (timerList.begin(), timerList.end(), timer);

	if (item != timerList.end()) {
		timer->setDistackModule (NULL);
		timerList.erase (item);
	}
}

cMessage* DistackOmnetModule::createTimerMsg (int timerid, string name)
{
	boost::mutex::scoped_lock scopedLock (sendMsgMutex);

	cMessage* scheduledMessage = new cMessage (name.length()>0 ? name.c_str() : NULL);
	scheduledMessage->setKind (timerid);

	return scheduledMessage;
}

void DistackOmnetModule::scheduleTimerMsg (simtime_t tm, cMessage* msg)
{
	boost::mutex::scoped_lock scopedLock (sendMsgMutex);
	scheduleAt (tm, msg);
}

#endif // OMNET_SIMULATION

//************************************************

LogOutputter* Distack::m_logoutputter = NULL;

Distack::Distack (string configfile)
{
	//
	// switch on memory debugging when debugging with msvc
	//

#if (defined (_DEBUG) & defined (_MSC_VER))
	////_crtBreakAlloc = 18687;
	_CrtSetDbgFlag (	_CRTDBG_ALLOC_MEM_DF	| 
						_CRTDBG_LEAK_CHECK_DF	| 
						_CRTDBG_CHECK_ALWAYS_DF	);
#endif 

	Configuration::init (configfile);
}

Distack::~Distack ()
{
}

#ifdef OMNET_SIMULATION

void Distack::setDistackModule (DistackOmnetModule* module)
{
	distackModule = module;
}

#endif // OMNET_SIMULATION

bool Distack::startup ()
{	
	//
	// configure log outputting
	//

	LogLevel::LOG_LEVEL logLevel = Configuration::instance()->getLogLevel();

	if (m_logoutputter == NULL) {
		m_logoutputter		= new StdOutLogOutputter ();
		Log::addLogOutput	(m_logoutputter);
		Log::setLogLevel	(logLevel);
	}
	
	//
	// startup all the stuff
	//

	m_networkmanager = new NetworkManager ();
	m_modulemanager  = new ModuleManager  ();

#ifdef OMNET_SIMULATION
	m_modulemanager->setDistackModule (distackModule);
#endif // OMNET_SIMULATION	 

	bool ret = true;

	ret &= m_networkmanager->startup ();	
	if (! ret) return false;

	//
	// start up the module manager and load the initial stage
	//

	m_modulemanager->startup ();

	//
	// start eating frames ... 
	//

#ifdef OMNET_SIMULATION
	m_modulemanager->setDistackModule (distackModule);
	NetworkInterfaceOmnetpp* omnetiface = (NetworkInterfaceOmnetpp*)m_networkmanager->getNetworkInterface ();
	omnetiface->setFrameDistribution (&(m_modulemanager->getFrameDistribution ()));
#endif // OMNET_SIMULATION	

	ret &= m_modulemanager->startFrameEating (m_networkmanager->getFrameSource ());
	if (! ret) return false;
	
	return ret;
}

bool Distack::shutdown ()
{
	bool ret = true;

	ret &= m_modulemanager->stopFrameEating	();
	ret &= m_networkmanager->shutdown		();
	
	delete m_networkmanager;
	delete m_modulemanager;

	if (m_logoutputter != NULL) {
		Log::removeLogOutput (m_logoutputter);
		delete m_logoutputter;
		m_logoutputter = NULL;
	}
	
	return ret;
}

NetworkManager*	Distack::getNetworkManager ()
{
	return m_networkmanager;
}

ModuleManager* Distack::getModuleManager ()
{
	return m_modulemanager;
}

//************************************************
