#include "NetworkInterfaceOmnetpp.h"
#include "base/module/FrameDistribution.h"

#ifdef OMNET_SIMULATION

NetworkInterfaceOmnetpp::NetworkInterfaceOmnetpp ()
:	NetworkInterface	("OmnetppInterface", NetworkInterface::CAPTURE_TYPE_ONLINE, true, (unsigned int) pow (2.0, 16)),
	frameDistribution	(NULL)
{
}

NetworkInterfaceOmnetpp::~NetworkInterfaceOmnetpp ()
{
	if (getState() == NetworkInterface::INTERFACE_STATE_OPEN)
		close ();
}

bool NetworkInterfaceOmnetpp::open ()
{
	return true;
}

bool NetworkInterfaceOmnetpp::close ()
{
	return true;
}

Frame* NetworkInterfaceOmnetpp::read ()
{
	return NULL;
}

void NetworkInterfaceOmnetpp::handleMessage (cMessage* msg)
{
	handleProtocolMessage (msg);
}

void NetworkInterfaceOmnetpp::setFrameDistribution (FrameDistribution* distrib)
{
	frameDistribution = distrib;
}

void NetworkInterfaceOmnetpp::handleProtocolMessage (cMessage* msg)
{
	if (msg == NULL) return;

	//
	// create the distack encapsulated Packets
	// according to the omnet packets ... 
	//

	Frame* frame = convertOmnetToDistack (msg);	
	delete msg;

	if (frame == NULL) return;

	frameDistribution->manualFrameInjection (frame);
}

Frame* NetworkInterfaceOmnetpp::convertOmnetToDistack (cMessage* message)
{
	cMessage* currentMessage = message;

	//
	// we do not use the full capacity of protocol encapsulation etc.
	// here. Maybe worth updating this code to support such stuff. 
	// Currently not needed ...
	//

	//
	// first protocol must be IP as we are an IPv4-Router
	//

	IPDatagram* ipDatagram = check_and_cast<IPDatagram*>(currentMessage);
	if(ipDatagram == NULL) return NULL;

	IpPacket*	ipPacket	= (IpPacket*) convertMessage (ipDatagram);
	Packet*		packetChain = ipPacket;
	Packet*		current		= ipPacket;

	//
	// handle IP-in-IP packets
	//

	if (ipPacket->getNextProtocol() == Packet::PROTO_IP) {

		currentMessage		 = currentMessage->encapsulatedMsg ();
		IPDatagram* datagram = check_and_cast<IPDatagram*> (currentMessage);

		Packet* innerIpPacket = convertMessage (datagram);

		ipPacket->setNextPacket	  (innerIpPacket);
		ipPacket->setNextProtocol (Packet::PROTO_IP);
		
		current = innerIpPacket;
	
	} // if (ipPacket->getProtocol() == Packet::PROTO_IP)
	
	//
	// handle the transport layer protocol
	//

	current->setNextProtocol (IpPacket::getTransportProtocol (((IpPacket*)current)->getProtocol()));

	Packet*			 transportPacket = NULL;
	Packet::PROTOCOL transportProto  = current->getNextProtocol ();
	currentMessage					 = currentMessage->encapsulatedMsg ();

	if (currentMessage != NULL) {

		switch (transportProto) {
			case Packet::PROTO_TCP:
				{
					TCPSegment* tcpSegment = check_and_cast<TCPSegment*>(currentMessage);
					transportPacket = convertMessage (tcpSegment);
					break;
				}
			case Packet::PROTO_UDP:
				{
					UDPPacket* udpPacket = check_and_cast<UDPPacket*>(currentMessage);
					transportPacket = convertMessage (udpPacket);
					break;	
				}
			case Packet::PROTO_ICMP:	
				{
					ICMPMessage* icmpMsg = check_and_cast<ICMPMessage*>(currentMessage);
					transportPacket = convertMessage (icmpMsg);
					break;
				}
			default:					
				{
					transportPacket = NULL;		
					break;
				}
		} //switch (transportProto)
	} // if (currentMessage != NULL)

	current->setNextPacket   (transportPacket);
	current->setNextProtocol (transportProto);

	//
	// no more protocols in upper layers as Omnet does not support any
	// (also no IP-in-ICMP etc. stuff is available, so we are done
	// but there might be some data which we model with the PayloadPacket
	//

	if (transportPacket != NULL) {

		cMessage* payloadMessage = currentMessage->encapsulatedMsg ();

		if (payloadMessage == NULL) {
			transportPacket->setNextPacket	 (NULL);
			transportPacket->setNextProtocol (Packet::PROTO_NONE);
		} else {
			PayloadPacket* payload = (PayloadPacket*)PacketFactory::instance()->createPacket (Packet::PROTO_DATA_PAYLOAD);
			
			payload->setData		 (NULL, 0);
			payload->setSize		 (payloadMessage->byteLength(), true);
			payload->setNextPacket	 (NULL);
			payload->setNextProtocol (Packet::PROTO_NONE);

			transportPacket->setNextPacket	 (payload);
			transportPacket->setNextProtocol (Packet::PROTO_DATA_PAYLOAD);
		}
	
	} // if (transportPacket != NULL)

	//
	// put it into a frame
	//

	Frame* frame = PacketFactory::instance()->createFrame ();
	frame->payloadpacket = packetChain;

	frame->capturelength = frame->payloadpacket->getSize  ();
	frame->packetlength  = frame->payloadpacket->getSize  ();
	
	//
	// set the timestamp for the frame
	//

	simtime_t timestamp = message->timestamp ();

	double intPart   = 0;
	double fractPart = 0;

	fractPart = modf (timestamp, &intPart);

	frame->timestamp.tv_sec  = (long)intPart;
	frame->timestamp.tv_usec = (long)fractPart * 1000;

	assert (frame->timestamp.tv_sec  >= 0);
	assert (frame->timestamp.tv_usec >= 0);

	return frame;
}

Packet* NetworkInterfaceOmnetpp::convertMessage (IPDatagram* msg)
{
	IpPacket* ipPacket = (IpPacket*)PacketFactory::instance()->createPacket (Packet::PROTO_IP);
	ipPacket->setSize (msg->byteLength(), true);

	ipPacket->setVersion		((unsigned char)msg->version());
	ipPacket->setHeaderlength	((unsigned char)msg->headerLength());
	ipPacket->setTos			(0); // no support in Omnet
	ipPacket->setTotallen		((short)msg->byteLength()); 
	ipPacket->setIdentification (msg->identification());
	ipPacket->setFlag			(IpPacket::IP_FLAG_R, false);
	ipPacket->setFlag			(IpPacket::IP_FLAG_DF, 	msg->dontFragment());
	ipPacket->setFlag			(IpPacket::IP_FLAG_MF, 	msg->moreFragments());
	ipPacket->setFragoffset		(msg->fragmentOffset());
	ipPacket->setTtl			((unsigned char)msg->timeToLive());
	ipPacket->setProtocol		(msg->transportProtocol());
	ipPacket->setChecksum		(0); // no support in Omnet
	ipPacket->setSourceip		(IP_ADDR (msg->srcAddress()));
	ipPacket->setDestip			(IP_ADDR (msg->destAddress()));
	ipPacket->setOptions		(NULL, 0); // tricky to implement

	return ipPacket;
}

Packet* NetworkInterfaceOmnetpp::convertMessage (TCPSegment* msg)
{
	TcpPacket* tcpPacket = (TcpPacket*)PacketFactory::instance()->createPacket (Packet::PROTO_TCP);
	tcpPacket->setSize (msg->byteLength(), true);

	tcpPacket->setSourceport	(msg->srcPort());
	tcpPacket->setDestport		(msg->destPort());
	tcpPacket->setSequencenum	(msg->sequenceNo());
	tcpPacket->setAcknum		(msg->ackNo());
	tcpPacket->setHeaderlength	(TCP_HEADER_NO_OPTIONS_LEN);
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_FIN, msg->finBit());
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_SYN, msg->synBit());
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_RST, msg->rstBit());
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_PSH, msg->pshBit());
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_ACK, msg->ackBit());
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_URG, msg->urgBit());
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_ECN, false);
	tcpPacket->setFlag			(TcpPacket::TCP_FLAG_CWR, false);
	tcpPacket->setWindowsize	((unsigned short)msg->window());
	tcpPacket->setChecksum		(0);
	tcpPacket->setUrgentpointer ((unsigned short) msg->urgentPointer());
	tcpPacket->setOptions		(NULL, 0); // tricky to implement

	return tcpPacket;
}

Packet* NetworkInterfaceOmnetpp::convertMessage (UDPPacket* msg)
{	
	UdpPacket* udpPacket = (UdpPacket*)PacketFactory::instance()->createPacket (Packet::PROTO_UDP);
	udpPacket->setSize (msg->byteLength(), true);

	udpPacket->setSourceport	(msg->sourcePort());
	udpPacket->setDestport		(msg->destinationPort());
	udpPacket->setLen			((short)msg->byteLength());
	udpPacket->setChecksum		(0); // no support from Omnet

	return udpPacket;
}

Packet* NetworkInterfaceOmnetpp::convertMessage (ICMPMessage* msg)
{
	IcmpPacket* icmpPacket = new IcmpPacket ();
	icmpPacket->setSize (msg->byteLength(), true);

	//
	// omnet uses different icmp type values. code is the same
	//

	IcmpPacket::ICMP_TYPE type = IcmpPacket::ICMP_TYPE (0);

	switch (msg->getType()) {
		case ICMP_DESTINATION_UNREACHABLE:	type = IcmpPacket::ICMP_TYPE_DEST_UNREACHABLE;	break;
		case ICMP_REDIRECT:					type = IcmpPacket::ICMP_TYPE_REDIRECT;			break;
		case ICMP_TIME_EXCEEDED:			type = IcmpPacket::ICMP_TYPE_TIME_EXCEEDED;		break;
		case ICMP_PARAMETER_PROBLEM:		type = IcmpPacket::ICMP_TYPE_PARAM_PROBLEM;		break;
		case ICMP_ECHO_REQUEST:				type = IcmpPacket::ICMP_TYPE_ECHO_REQUEST;		break;
		case ICMP_ECHO_REPLY:				type = IcmpPacket::ICMP_TYPE_ECHO_REPLY;		break;
		case ICMP_TIMESTAMP_REQUEST:		type = IcmpPacket::ICMP_TYPE_TIMESTAMP_REQUEST;	break;
		case ICMP_TIMESTAMP_REPLY:			type = IcmpPacket::ICMP_TYPE_TIMESTAMP_REPLY;	break;
	}

	icmpPacket->setType			(type);
	icmpPacket->setCode			(msg->getCode());
	icmpPacket->setChecksum		(0); // no support from Omnet
	icmpPacket->setMisc			(0); // no support from Omnet

	return icmpPacket;
}

#endif // OMNET_SIMULATION
